﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading.Tasks;

namespace ConsoleApp1
{
    class Program
    {
        static void Main(string[] args)
        {
            IPv4Tables table = new IPv4Tables();
            Console.WriteLine(table.BinarySearch(IPAddress.Parse("103.60.16.16")).NationCode);
            IPAddress ipaddr;

            return;
        }

        public enum SearchMode
        {
            Sequential,
            Binary,
            Parallel,
        }

        private static void CheckAddress(IPv4Tables table, List<IPAddress> randomAddresses, SearchMode mode)
        {
            List<IPAddress> publics = new List<IPAddress>();

            Stopwatch st = new Stopwatch();
            st.Start();

            switch (mode)
            {
                case SearchMode.Parallel:
                    Parallel.ForEach(randomAddresses, (item) =>
                    {
                        var found = table.BinarySearch(item);
                    });
                    break;

                case SearchMode.Sequential:
                    foreach (IPAddress addr in randomAddresses)
                    {
                        var found = table.Search(addr);
                        Console.WriteLine(found.NationCode);
                    }
                    break;

                case SearchMode.Binary:
                    foreach (IPAddress addr in randomAddresses)
                    {
                        var found = table.BinarySearch(addr);
                        Console.WriteLine(found.NationCode);
                    }
                    break;
            }

            st.Stop();
            Console.WriteLine($"{st.ElapsedMilliseconds}");
        }

        private static List<IPAddress> CreateTestAddresses(int count, int seed = 0)
        {
            if (seed == 0)
            {
                seed = Environment.TickCount;
            }

            Random rand = new Random(seed);

            byte[] buffer = new byte[4];
            List<IPAddress> addrs = new List<IPAddress>();

            for (int i = 0; i < count; i++)
            {
                rand.NextBytes(buffer);
                IPAddress ipAddr = new IPAddress(buffer);

                addrs.Add(ipAddr);
            }

            return addrs;
        }
    }

    public static class IPAddressHelper
    {
        public static uint ToUInt32(this IPAddress addr)
        {
            // byte[] buf = addr.GetAddressBytes();
            uint dstValue = 0;

            /*
            {
                return BitConverter.ToUInt32(buf.Reverse().ToArray(), 0);
            }

            {
                return (uint)(buf[0] << 24 | buf[1] << 16 | buf[2] << 8 | buf[3]);
            }
            */

            // uint srcValue = (uint)addr.Address;
            uint srcValue = (uint)addr.GetHashCode();

            unsafe
            {
                byte* pDstValue = (byte*)&dstValue;
                byte* pSrcValue = (byte*)&srcValue;

                *(pDstValue + 0) = *(pSrcValue + 3);
                *(pDstValue + 1) = *(pSrcValue + 2);
                *(pDstValue + 2) = *(pSrcValue + 1);
                *(pDstValue + 3) = *(pSrcValue + 0);

                /*
                *(pDstValue + 0) = (byte)((srcValue & 0xFF000000) >> 24);
                *(pDstValue + 1) = (byte)((srcValue & 0x00FF0000) >> 16);
                *(pDstValue + 2) = (byte)((srcValue & 0x0000FF00) >> 8);
                *(pDstValue + 3) = (byte)((srcValue & 0x000000FF) >> 0);
                */

                return dstValue;
            }
        }
    }

    public class IPv4Range : IComparable<IPv4Range>
    {
        public DateTime Created;
        public string NationCode;
        public IPAddress Started;
        public IPAddress Ended;
        public string Prefix;
        public DateTime Allocated;

        public int CompareTo(IPv4Range other)
        {
            uint otherAddr = other.Started.ToUInt32();
            uint thisAddr = Started.ToUInt32();

            if (otherAddr == thisAddr)
            {
                return 0;
            }

            if (otherAddr > thisAddr)
            {
                return -1;
            }

            return 1;
        }

        public bool IsInRange(IPAddress target)
        {
            uint started = Started.ToUInt32();
            uint ended = Ended.ToUInt32();

            uint current = target.ToUInt32();

            if (started <= current && current <= ended)
            {
                return true;
            }

            return false;
        }

        public override string ToString()
        {
            return $"{NationCode}({Started}~{Ended})";
        }
    }

    public class IPv4Tables
    {
        public static IPv4Range[] _ranges;

        static IPv4Tables()
        {
            _ranges = ReadTables();
            Array.Sort(_ranges);
        }

        private static IPv4Range[] ReadTables()
        {
            List<IPv4Range> list = new List<IPv4Range>();

            foreach (string line in File.ReadAllLines("ipv4.csv").Skip(1))
            {
                string[] items = line.Split(',');

                IPv4Range range = new IPv4Range();
                range.Created = DateTime.ParseExact(items[0], "yyyyMMdd", CultureInfo.InvariantCulture);
                range.NationCode = items[1];
                range.Started = IPAddress.Parse(items[2]);
                range.Ended = IPAddress.Parse(items[3]);
                range.Prefix = items[4];
                range.Allocated = DateTime.ParseExact(items[5], "yyyyMMdd", CultureInfo.InvariantCulture);

                list.Add(range);
            }

            return list.ToArray();
        }

        public IPv4Range BinarySearch(IPAddress item)
        {
            IPAddressComparer comparer = new IPAddressComparer();
            int n = Array.BinarySearch(_ranges, item, comparer);
            if (n < 0)
            {
                return null;
            }

            return _ranges[n];
        }

        public class IPAddressComparer : System.Collections.IComparer
        {
            public int Compare(object x, object y)
            {
                IPv4Range range = x as IPv4Range;
                IPAddress addr = y as IPAddress;

                uint started = range.Started.ToUInt32();
                uint ended = range.Ended.ToUInt32();
                uint target = addr.ToUInt32();

                if (started <= target && target <= ended)
                {
                    return 0;
                }

                if (started > target)
                {
                    return 1;
                }

                return -1;
            }
        }

        public IPv4Range Search(IPAddress item)
        {
            foreach (IPv4Range range in _ranges)
            {
                if (range.IsInRange(item) == true)
                {
                    return range;
                }
            }

            return null;
        }
    }
}
